# File svdfit/Src/eigenfit.x
# February 5, 2009
# By Doug Mink, Harvard-Smithsonian Center for Astrophysics
 
# EIGENFIT fits a set of eigenvectors to a spectrum and returns the residuals

# Copyright(c) 2000-2009 Smithsonian Astrophysical Observatory
# You may do anything you like with this file except remove this copyright.
# The Smithsonian Astrophysical Observatory makes no representations about
# the suitability of this software for any purpose.  It is provided "as is"
# without express or implied warranty.
 
include	<imhdr.h>
include	<imio.h>
include	<fset.h>
include <smw.h>
include "rvsao.h"
include	"xcv.h"
include "contin.h"

procedure eigenfit (specfile,specdir,specap,specband,nbin,fitmax,
		    rebspec,resspec,aptrack,fitspec,rebwl,reblines,rebcont,
		    ispec,nspec,imout,outpath,cfd, verbose)

char	specfile[ARB]	# Object spectrum file name
char	specdir[ARB]	# Directory for object spectra
int	specap		# Object aperture to read from multispec file
int	specband	# Object band to read from multispec file
int	nbin		# Number of pixels in rebinned spectrum and eigenvectors
double	fitmax		# Maximum flux ratio for single vector fit
pointer	rebspec		# Rebinned spectrum (filled and returned)
pointer	resspec		# Spectrum residuals (filled and returned)
bool	aptrack		# If true, eigenvector aperture tracks object aperture
pointer	fitspec		# Fit eigenvectors (filled and returned)
pointer	rebwl		# Wavelength vector for rebinned spectra
pointer	reblines	# Lines removed from rebinned spectra
pointer	rebcont		# Continuum vector for rebinned spectra
int	ispec		# Number of this spectrum
int	nspec		# Number of spectra in output file
pointer	imout		# Output data structure, NULL on first spectrum per file
char	outpath[ARB]	# Output file name
int	cfd		# File descripter for coefficient file
bool	verbose		# Print results to log if true

pointer spectrum	# Object spectrum
pointer specim		# Object image header structure

int	i, j
char	specpath[SZ_PATHNAME]	# Object spectrum path name
char	vecfiles[SZ_PATHNAME]	# List of eigenvector spectra
char	vecfile[SZ_PATHNAME]	# Eigenvector spectrum file name
char	vecpath[SZ_PATHNAME]	# Eigenvector spectrum path name
char	vecdir[SZ_PATHNAME]	# Directory for eigenvector spectra
char	vecnums0[SZ_LINE]	# List of multispec eigenvector spectra to read
char	vecnums[SZ_LINE]	# List of multispec eigenvector spectra to read
int	rmode			# Report format (1=normal,2=one-line)
char	xlab[SZ_LINE]		# Title for wavelength plots of spectrum

pointer	specwcs		# Pointer to first data point with WCS
pointer	wlwcs		# Pointer to first wavelength with WCS
int	npspec		# Number of spectrum points with WCS
pointer	vecwcs		# Pointer to first eigenvector pixel with WCS
pointer	twlwcs		# Pointer to first eigenvector wavelength with WCS
int	nvtemp		# Number of eigenvector points with WCS
int	npout		# Number of pixels in residual file
double	pixshift0	# Pixel shift fore call to rebin()

double	facnorm		# Renormalize by dividing by mean/this
int	vspec		# Eigenvector spectrum to read from multispec file
bool	pltres

pointer	veclist		# List of eigenvector files

double 	minwav0,maxwav0	# Starting,ending wavelengths from parameter file
double 	minwav1,maxwav1	# Spectrum and eigenvector overlap wavelength limits
double 	minwav,maxwav	# Wavelength limits for cross-correlation
double	dwav		# Wavelength per pixel
double	spwl1,spwl2	# Spectrum first and last wavelengths in angstroms
double	tmwl1,tmwl2	# Eigenvector first and last wavelengths in angstroms
double	tmdwl		# Eigenvector angstroms per pixel
int	nvf		# Number of eigenvector actually fit
int	vband		# Multispec band for eigenvectors
double	smean, vmean	# Mean flux of spectrum and eigenvector in counts/pixel
real 	spavg, speci
double 	sptot
int	ivec, nsp
char	keyword[16]
char	tstring[SZ_LINE]
bool	isvec

pointer vecim		# Eigenvector image header structure
pointer	vecsh		# Eigenvector spectrum header structure
char	title[SZ_LINE]

int	nvspec,nvspec0	# Number of eigenvector multispec spectra
int	vspec_range[3,MAX_RANGES]
int	ip,jp,lfile	# Limits for multispec aperture decoding
char	lbracket[3]	# left bracket, parenthesis, brace
char	rbracket[3]	# right bracket, parenthesis, brace
int	tpix1, tpix2, spix1, spix2
double	dindef
double	tmshft, spshft
char	cdot
double	chi, chisq
double	coeff[MAXTEMPS]
double	oldcoeff
bool	addback		# Add continuum back after removing eigenvectors
bool	spwrite
 
bool	clgetb()
int	clgeti()
real	max(), min()
double	clgetd()
int	stridx(), stridxs()
int	decode_ranges(),get_next_number()
int	imtgetim(), imaccess(), strlen(), ldir
pointer	imtopen()
double	wcs_p2w()

include	"rvsao.com"
include "results.com"
include "contin.com"

define	newtemp_ 40
define	newtap_	 50
define	nextemp_ 60
define	endtemp_ 80
define	endfit_  90
 
begin
	c0 = 299792.5d0
	pixshift0 = 0.d0
	cdot = '.'
	schop = FALSE
	tchop = FALSE
	tmshft = 0.d0
	dindef = INDEFD
	specsh = NULL
	spwrite = FALSE
	vecsh = NULL
	call sprintf (lbracket,3,"[({")
	call sprintf (rbracket,3,"])}")

#	if (debug) {
#	    call printf ("EIGENFIT: nbin = %d, nspec = %d\n")
#		call pargi (nbin)
#		call pargi (nspec)
#	    }

# Load spectrum
	ldir = strlen (specdir)
	if (ldir > 0) {
	    if (specdir[ldir] != '/') {
		specdir[ldir+1] = '/'
		specdir[ldir+2] = EOS
		}
	    call strcpy (specdir,specpath,SZ_PATHNAME)
	    call strcat (specfile,specpath,SZ_PATHNAME)
	    }
	else
	    call strcpy (specfile,specpath,SZ_PATHNAME)

# Check for readability of object spectrum
	if (imaccess (specpath, READ_ONLY) == NO) {
	    call eprintf ("EIGENFIT: cannot read spectrum path %s \n")
		call pargstr (specpath)
	    return
	    }

#	debug = FALSE
	call getspec (specpath, specap, specband, spectrum, specim, spwrite)
	if (specim == ERR) {
	   call eprintf ("EIGENFIT: Error reading spectrum %s\n")
	 	call pargstr (specpath)
	    go to endfit_
	    }
	if (specband > 0) {
	    call sprintf (specid, SZ_PATHNAME,"%s[%d]")
		call pargstr (specpath)
		call pargi (specband)
	    }
	else if (specap > 0) {
	    call sprintf (specid, SZ_PATHNAME,"%s[%d]")
		call pargstr (specpath)
		call pargi (specap)
	    }
	else {
	    call sprintf (specid, SZ_PATHNAME, "%s")
		call pargstr (specpath)
	    }

	call wcs_set (specsh)
	spix1 = NP1(specsh)
	spix2 = NP2(specsh)
	spwl1 = wcs_p2w (double (spix1))
	spwl2 = wcs_p2w (double (spix2))
	spdwl = (spwl2 - spwl1) / double (spix2 - spix1)

	call strcpy (LABEL(specsh), xlab, SZ_LINE)
	if (strlen (xlab) <= 0)
	    call strcpy ("Wavelength", xlab, SZ_LINE)
	if (strlen (UNITS(specsh)) > 0) {
	    call strcat (" in ",xlab,SZ_LINE)
	    call strcat (UNITS(specsh),xlab,SZ_LINE)
	    }
	else
	    call strcat (" in Angstroms",xlab,SZ_LINE)

# Renormalize spectrum if requested
	facnorm = clgetd ("renormalize")
	if (facnorm != 0.0) {
	    sptot = 0.d0
	    nsp = 0
	    do i = 1, specpix {
		speci = Memr[spectrum+i-1]
		if (speci != 0.) {
		    sptot = sptot + speci
		    nsp = nsp + 1
		    }
		}
	    if (nsp > 0) {
		spavg = (sptot / double (nsp)) / facnorm
		do i = 1, specpix {
		    if (Memr[spectrum+i-1] != 0.) {
			Memr[spectrum+i-1] = Memr[spectrum+i-1] / spavg
			}
		    }
		}
	    else {
		call eprintf ("EIGENFIT: Spectrum is all zeroes\n")
		go to endfit_
		}
	    }

# Set pixel limits of spectrum WCS
	specwcs = spectrum + NP1(specsh) - 1
	wlwcs = SX(specsh) + NP1(specsh) - 1
	npspec = NP2(specsh) - NP1(specsh) + 1

# Set number of bins if not already set
	if (nbin == INDEF || nbin < 1) {
	    npout = npspec

# Allocate rebinned, wavelength, residual, and sky spectrum buffers
	    if (rebspec != NULL) {
		call mfree (rebspec, TY_REAL)
		call mfree (rebwl, TY_REAL)
		call mfree (resspec, TY_REAL)
		call mfree (fitspec, TY_REAL)
		}
	    call malloc (rebspec, npout, TY_REAL)
	    call malloc (rebwl, npout, TY_REAL)
	    call malloc (resspec, npout, TY_REAL)
	    call malloc (fitspec, npout, TY_REAL)
	    nbin = npout
	    }
	else
	    npout = nbin

# Buffer vectors for continuum removal
	if (contvec == NULL && conproc != NOCONT) {
	    call malloc (contvec, npout, TY_REAL)
	    call malloc (contbuf, npout, TY_REAL)
	    call malloc (contwts, npout, TY_REAL)
	    call malloc (contrej, npout, TY_REAL)
	    }

# If plot enabled, show the object spectrum.
	if (pltspec) {
	    if (nspec > 0) {
		call sprintf (title, SZ_LINE, "%d: ")
		    call pargi (ispec)
		}
	    call sprintf (title, SZ_LINE, "%s original")
		call pargstr (specid)
	    call plotspec (npspec,Memr[specwcs],title,Memr[wlwcs],xlab,nsmooth)
	    }

#	if (debug) {
#	    call printf ("EIGENFIT: nbin = %d -> %d title= %s\n")
#		call pargi (npspec)
#		call pargi (nbin)
#		call pargstr (title)
#	    }

# Optional wavelength limits for eigenvector fit
	minwav0 = clgetd ("st_lambda")
	if (minwav0 == 0.d0)
	    minwav0 = dindef
	maxwav0 = clgetd ("end_lambda")
	if (maxwav0 == 0.d0)
	    maxwav0 = dindef

# Eignenvectors to which to fit spectra
	call clgstr ("eigenfiles",vecfiles,SZ_PATHNAME)
	call clgstr ("eigendir",vecdir,SZ_PATHNAME)
	ldir = strlen (vecdir)
	if (ldir > 0 && vecdir[ldir] != '/' && vecdir[ldir] != '$') {
	    vecdir[ldir+1] = '/'
	    vecdir[ldir+2] = EOS
	    }

# Band (object, raw, sky) of eigenvector spectrum to use
	vband = clgeti ("eigenband")

# Multispec eigenvector numbers (use only first if multiple files)
	call clgstr ("eigenum",vecnums0,SZ_LINE)
	if (decode_ranges (vecnums0, vspec_range, MAX_RANGES, nvspec0) == ERR)
	    call error (1, "EIGENFIT: Illegal multispec list")

# Load all eigenvectors
	ivec = 0
	nvf = 0
	if (debug) {
	    call printf ("EIGENFIT: nctemp = %d\n")
		call pargi (nctemp)
	    }

	if (nctemp == 0) {
	    ntemp = 0
	    veclist = imtopen (vecfiles)
	    }

# Initialize new eigenvector file
newtemp_
	if (nctemp > 0) go to nextemp_
	vspec = -1

# Get next eigenvector spectrum file name from the list
	if (imtgetim (veclist, vecfile, SZ_PATHNAME) == EOF)
	    go to endtemp_

# Check for specified apertures in multispec eigenvector file
	ip = stridxs (lbracket,vecfile)
	if (ip > 0) {
	    lfile = strlen (vecfile)
	    vecfile[ip] = EOS
	    ip = ip + 1
	    jp = 1
	    while (stridx (vecfile[ip],rbracket) == 0 && ip <= lfile) {
		vecnums[jp] = vecfile[ip]
		vecfile[ip] = EOS
		ip = ip + 1
		jp = jp + 1
		}
	    vecnums[jp] = EOS
	    if (decode_ranges (vecnums,vspec_range,MAX_RANGES,nvspec) == ERR)
		call error (1, "EIGENFIT: Illegal eigenvector multispec list")
	    }
	else if (aptrack) {
	    call sprintf (vecnums, "%d")
		call pargi (specap)
	    if (decode_ranges (vecnums,vspec_range,MAX_RANGES,nvspec) == ERR)
		call error (1, "EIGENFIT: Illegal eigenvector multispec list")
	    }
	else {
	    call strcpy (vecnums0,vecnums,SZ_LINE)
	    if (decode_ranges (vecnums,vspec_range,MAX_RANGES,nvspec) == ERR)
		call error (1, "EIGENFIT: Illegal eigenvector multispec list")
	    }
#	if (debug) {
#	    call printf ("EIGENFIT: next eigenvector file is %s[%s] = %d aps\n")
#		call pargstr (vecfile)
#		call pargstr (vecnums)
#		call pargi (nvspec)
#	    }

# Check for readability of eigenvector spectrum
	if (stridx ("/",vecfile) == 1 || stridx ("$",vecfile) > 0 ||
	    stridx (cdot,vecfile) == 1)
	    call strcpy (vecfile,vecpath,SZ_PATHNAME)
	else {
	    call strcpy (vecdir,vecpath,SZ_PATHNAME)
	    call strcat (vecfile,vecpath,SZ_PATHNAME)
	    }
	if (imaccess (vecpath, READ_ONLY) == NO) {
	    call eprintf ("EIGENFIT: cannot read eigenvector file %s \n")
		call pargstr (vecpath)
	    go to newtemp_
	    }

# Get next multispec aperture number from list
newtap_
	if (get_next_number (vspec_range, vspec) == EOF)
	    go to newtemp_

# Start new eigenvector here if it has been cached
nextemp_
	ivec = ivec + 1
	if (ntemp > 0 && ivec > ntemp ) 
	    go to endtemp_

# Use saved eigenvector spectrum
	if (nctemp > 0 && ivec <= nctemp) {
	    vecsh = tmpsh[ivec]
	    vecim = tmpim[ivec]
	    vspec = tmpspec[ivec]
#	    if (debug) {
#		call printf ("EIGENFIT: Vector %d from cache\n")
#		    call pargi (ivec)
#		}
	    }

# Load eigenvector spectrum
	else {
	    call getvec (vecfile,vspec,vband,vecdir,vecim,vecsh,ivec)
	    if (vecim == ERR)
		go to newtap_

	    if (vband > 0) {
		call sprintf (tempid[1,ivec], SZ_PATHNAME,"%s[%d]")
		    call pargstr (vecfile)
		    call pargi (vband)
		}
	    else if (vspec > 0) {
		call sprintf (tempid[1,ivec], SZ_PATHNAME,"%s[%d]")
		    call pargstr (vecfile)
		    call pargi (vspec)
		}
	    else {
		call sprintf (tempid[1,ivec], SZ_PATHNAME, "%s")
		    call pargstr (vecfile)
		}
	    if (ivec < MAXTEMPS) {
		tmpsh[ivec] = vecsh
		tmpim[ivec] = vecim
		tmpspec[ivec] = vspec
		}
	    if (plttemp) {
		tpix1 = NP1(vecsh)
		tpix2 = NP2(vecsh)
		vecwcs = SY(vecsh) + tpix1 - 1
		twlwcs = SX(vecsh) + tpix1 - 1
		nvtemp = tpix2 - tpix1 + 1
		call sprintf (title, SZ_LINE, "%d: %s eigenvector original")
		    call pargi (tmpspec[ivec])
		    call pargstr (tempid[1,ivec])
		call plotspec (nvtemp,Memr[vecwcs],title,Memr[twlwcs],xlab,
			       nsmooth)
		}
	    }
	nvf = nvf + 1

# Set pixel limits of eigenvector WCS and plot eigenvector, if requested
	tpix1 = NP1(vecsh)
	tpix2 = NP2(vecsh)
	vecwcs = SY(vecsh) + tpix1 - 1
	twlwcs = SX(vecsh) + tpix1 - 1
	nvtemp = tpix2 - tpix1 + 1

# Ignore wavelength limit parameters and use file limits
	overlap = FALSE
	call imgbpar (vecim,"OVERLAP",overlap)
	if (minwav0 == dindef && maxwav0 == dindef)
	    overlap = FALSE
#	if (debug) {
#	    if (overlap)
#		call printf ("EIGENFIT: Use entire overlapping wavelength space\n")
#	    else
#		call printf ("EIGENFIT: Use wavelength limit parameters, if set\n")
#	    }
	toverlap[ivec] = overlap

	call wcs_set (vecsh)
	tmwl1 = wcs_p2w (double (tpix1))
	tmwl2 = wcs_p2w (double (tpix2))
	tmdwl = (tmwl2 - tmwl1) / double (nvtemp - 1)
	if (debug) {
	    call printf ("EIGENFIT: Eigenvector %.4f to %.4f by %.7f = %d\n")
		call pargd (tmwl1)
		call pargd (tmwl2)
		call pargd (tmdwl)
		call pargi (nvtemp)
	    }

	# Check to make sure that eigenvector and spectrum overlap
	if (spwl1 > tmwl2) {
	    call eprintf ("EIGENFIT: Spectrum > %9.2f, eigenvector < %9.2f angstroms\n")
		call pargd (spwl1)
		call pargd (tmwl2)
	    go to endfit_
	    }
	if (spwl2 < tmwl1) {
	    call eprintf ("EIGENFIT: Spectrum < %9.2f, eigenvector > %9.2f angstroms\n")
		call pargd (spwl2)
		call pargd (tmwl1)
	    go to endfit_
	    }

	# Figure lambda and log-lambda limits and increment
	minwav1 = max (spwl1, tmwl1)
	if ((overlap) || (minwav0==dindef) || (minwav0<minwav1))
	    minwav = minwav1
	else
	    minwav = minwav0

	maxwav1 = min (spwl2, tmwl2)
	if ((overlap) || (maxwav0==dindef) || (maxwav0>maxwav1))
	    maxwav = maxwav1
	else
	    maxwav = maxwav0
	dwav = (maxwav - minwav) / double (npout - 1)

	# Set up wavelength vector for intermediate graphs
	if (ivec == 1 || minwav != twl1[ivec-1] || maxwav != twl2[ivec-1]) {
	    do j = 1, npout {
		Memr[rebwl+j-1] = minwav + (double (j - 1) * dwav)
		}
	    }

	# If range is different for rebinning same eigenvector, do it again
	if (nctemp == 0 || (minwav != twl1[ivec] && maxwav != twl2[ivec])) {
	    twl1[ivec] = minwav
	    twl2[ivec] = maxwav
	    dwl[ivec] = dwav

	    #  Rebin eigenvector to overlap this spectrum
	    if (debug) {
		call printf("EIGENFIT: spectrum %9.7f to %9.7f by %.7f = %d\n")
		    call pargd (minwav)
		    call pargd (maxwav)
		    call pargd (dwav)
		    call pargi (npout)
		}
	    if (rebvec[ivec] != NULL)
		call mfree (rebvec[ivec], TY_REAL)
	    call malloc (rebvec[ivec], npout+2, TY_REAL)
	    if (tmwl1 == minwav && tmdwl == dwav && nvtemp == npout) {
		do i = 1, npout {
		    Memr[rebvec[ivec]+i-1] = Memr[vecwcs+i-1]
		    }
		}
	    else {
		call rebin (Memr[vecwcs], vecsh, tmshft,
			Memr[rebvec[ivec]], npout, minwav, dwl[ivec], pixshift0)
		}
#	    if (debug) {
#		call printf ("EIGENFIT: In  Lambda %f %f %f\n")
#		    call pargr (Memr[twlwcs])
#		    call pargr (Memr[twlwcs+1])
#		    call pargr (Memr[twlwcs+2])
#		call printf ("EIGENFIT: In  Eigenvector %f %f %f\n")
#		    call pargr (Memr[vecwcs])
#		    call pargr (Memr[vecwcs+1])
#		    call pargr (Memr[vecwcs+2])
#		call printf ("EIGENFIT: Out Lambda %f %f %f\n")
#		    call pargr (Memr[rebwl])
#		    call pargr (Memr[rebwl+1])
#		    call pargr (Memr[rebwl+2])
#		call printf ("EIGENFIT: Out Eigenvector %f %f %f\n")
#		    call pargr (Memr[rebvec[ivec]])
#		    call pargr (Memr[rebvec[ivec]+1])
#		    call pargr (Memr[rebvec[ivec]+2])
#		}
	    if (plttemp) {
		call sprintf (title, SZ_LINE, "%d: %s eigenvector rebinned\n")
		    call pargi (tmpspec[ivec])
		    call pargstr (tempid[1,ivec])
		call plotspec (npout,Memr[rebvec[ivec]],title,Memr[rebwl],xlab,nsmooth)
		}
	    }
	else if (plttemp) {
	    call sprintf (title, SZ_LINE, "%d: %s eigenvector from cache\n")
		call pargi (tmpspec[ivec])
		call pargstr (tempid[1,ivec])
	    call plotspec (npout,Memr[rebvec[ivec]],title,Memr[rebwl],xlab,nsmooth)
	    }

	# Allocate and initialize continuum line vector for current eigenvector
	if (cnvec[ivec] == NULL)
	    call malloc (cnvec[ivec], npout+2, TY_REAL)
	do i = 0, npout-1 {
	    Memr[cnvec[ivec]+i] = 0.0
	    }

	# Allocate and initialize removed line vector for current eigenvector
	if (lnvec[ivec] == NULL)
	    call malloc (lnvec[ivec], npout+2, TY_REAL)
	do i = 0, npout-1 {
	    Memr[lnvec[ivec]+i] = 0.0
	    }
	if (debug) {
	    call printf ("EIGENFIT: cnvec[%d][%d] = %f\n")
		call pargi (ivec)
		call pargi (npout)
		call pargr (Memr[cnvec[ivec]+npout-1])
	    call flush (STDOUT)
	    }

	# Subtract continuum and chop unwanted lines from eigenvector using ICFIT
	isvec = TRUE
	call eigencon (npout,Memr[rebvec[ivec]],Memr[rebwl],tempid[1,ivec],
		       nsmooth,ispec,nspec,isvec,Memr[lnvec[ivec]],Memr[cnvec[ivec]], vmean)

	# Reset per-eigenvector pointers
	vecim = NULL
	vecsh = NULL
	nvspec = nvspec - 1
	if (nvspec > 0)
	    go to newtap_
	go to newtemp_

#  Process spectrum
endtemp_
	if (debug) {
	    call printf("EIGENFIT: Overlap %9.4f to %9.4f by %9.4f, nbin = %d \n")
		call pargd (minwav)
		call pargd (maxwav)
		call pargd (dwav)
		call pargi (npout)
	    call printf("EIGENFIT: Spectrum %9.4f to %9.4f by %9.4f = %d %d\n")
		call pargd (spwl1)
		call pargd (spwl2)
		call pargd (spdwl)
		call pargi (SN(specsh))
		call pargi (DC(specsh))
	    }

	#  Rebin spectrum to overlap this eigenvector
	spshft = 0.d0
	if (tmwl1 == minwav && tmdwl == dwav && nvtemp == npout)
	    call amovr (Memr[spectrum],Memr[rebspec],npout)
	else
	    call rebin (Memr[spectrum],specsh, spshft,Memr[rebspec],npout,minwav,dwav,pixshift0)
#	if (debug) {
#	    call printf ("EIGENFIT: In  Lambda %f %f %f\n")
#		call pargr (Memr[SX(specsh)])
#		call pargr (Memr[SX(specsh)+1])
#		call pargr (Memr[SX(specsh)+2])
#	    call printf ("EIGENFIT: In  Spectrum  %f %f %f\n")
#		call pargr (Memr[SY(specsh)])
#		call pargr (Memr[SY(specsh)+1])
#		call pargr (Memr[SY(specsh)+2])
#	    call printf ("EIGENFIT: Out Lambda %f %f %f\n")
#		call pargr (Memr[rebwl])
#		call pargr (Memr[rebwl+1])
#		call pargr (Memr[rebwl+2])
#	    call printf ("EIGENFIT: Out Spectrum %f %f %f\n")
#		call pargr (Memr[rebspec])
#		call pargr (Memr[rebspec+1])
#		call pargr (Memr[rebspec+2])
#	    call flush (STDOUT)
#	    }

	# Initialize spectrum vector which will be filled with residuals
	do i = 0, npout-1 {
	    Memr[resspec+i] = Memr[rebspec+i]
	    }

	# Subtract continuum and chop unwanted lines from spectrum using ICFIT
	isvec = FALSE
	call eigencon (npout,Memr[resspec],Memr[rebwl],specname,nsmooth,
		       ispec,nspec,isvec,Memr[reblines],Memr[rebcont], smean)

	# Fit eigenvectors to this spectrum
	call fitvecs (Memr[resspec], npout, rebvec, nvf, coeff, chi, chisq, debug)

	# If coefficient is higher that it is supposed to be, use 1.0
	if (nvf == 1 && coeff[1] > fitmax) {
	    if (verbose) {
		do i = 1, nlogfd {
		    call fprintf (logfd[i], "File %s: coeff = %4f > %.4f\n")
			call pargstr (specid)
			call pargd (coeff[1])
			call pargd (fitmax)
		    }
		}

	    # Initialize spectrum vector which will be filled with residuals
	    do i = 0, npout-1 {
		Memr[resspec+i] = Memr[rebspec+i]
		}

	    # Subtract continuum and chop unwanted lines from spectrum
	    isvec = FALSE
	    call eigencon (npout,Memr[resspec],Memr[rebwl],specname,nsmooth,
			   ispec,nspec,isvec,Memr[reblines],Memr[rebcont], smean)
	    oldcoeff = coeff[1]
	    coeff[1] = 1.0
	    do i = 1, npout
		Memr[resspec+i-1] = Memr[resspec+i-1] - Memr[rebvec[1]+i-1]
	    }
	else
	    oldcoeff = 0.0

	if (verbose) {
	    do i = 1, nlogfd {
		if (nvf > 1) {
		    call fprintf (logfd[i], "File %s fit with %d eigenvectors\n")
			call pargstr (specid)
			call pargi (nvf)
		    }
		else {
		    call fprintf (logfd[i], "File %s fit by eigenvector * %.4f\n")
			call pargstr (specid)
			call pargd (coeff[1])
		    }
		}
	    }
	if (cfd > -1) {
	    if (ispec == 1) {
		call fprintf (cfd, "i    	ap   	chi       	chi2      ")
		do i = 1, nvf {
		    call fprintf (cfd, "	coeff%d")
			call pargi (i)
		    }
		call fprintf (cfd, "\n")
		call fprintf (cfd, "-----	-----	---------	---------")
		do i = 1, nvf {
		    call fprintf (cfd, "	------")
		    }
		call fprintf (cfd, "\n")
		}
	    call fprintf (cfd,"%5d	%5d	%9.2f	%9.2f")
		call pargi (ispec)
		call pargi (specap)
		call pargd (chi)
		call pargd (chisq)
	    do i = 1, nvf {
		call fprintf (cfd,"	%6.3f")
		    call pargd (coeff[i])
		}
	    call fprintf (cfd,"\n")
	    }

# Add back continuum if requested
	addback = FALSE
	addback = clgetb ("cont_add")
	if (addback) {

# If continuum removed from eigenvector(s), subtract it from spectrum, too
	    if (convec != NOCONT) {
		do i = 0, npout-1 {
		    do j = 1, nvf {
			Memr[rebcont+i] = Memr[rebcont+i] - (coeff[j]*Memr[cnvec[j]+i])
			Memr[reblines+i] = Memr[reblines+i] - (coeff[j]*Memr[lnvec[j]+i])
			Memr[rebvec[j]+i] = Memr[rebvec[j]+i] + Memr[cnvec[j]+i]
			}
		    }
		}

	    do i = 0, npout-1 {
		Memr[resspec+i] = Memr[resspec+i] + Memr[rebcont+i] + Memr[reblines+i]
		}
	    }
	if (debug) {
	    call printf ("EIGENFIT: Continuum - background and lines added back\n")
	    call flush (STDOUT)
	    }

# Close the eigenvector list for this object spectrum
	if (nctemp == 0) {
	    ntemp = ivec
	    call imtclose (veclist)
	    }
	if (ntemp <= MAXTEMPS)
	    nctemp = ntemp
#	if (debug) {
#	    call printf ("EIGENFIT: ntemp = %d nctemp = %d\n")
#		call pargi (ntemp)
#		call pargi (nctemp)
#	    }

# Save results for this object spectrum if at least one eigenvector fit
	if (nvf > 0) {
	    pltres  = clgetb ("res_plot")
	    if (pltres) {
		if (nspec < 0) {
		    call sprintf (title, SZ_LINE, "%s raw spectrum")
			call pargstr (specname)
		    call plotspec (npout,Memr[rebspec],title,Memr[rebwl],xlab,nsmooth)
		    call sprintf (title, SZ_LINE, "%s first eigenvector (sky)")
			call pargstr (specname)
		    call plotspec (npout,Memr[rebvec[1]],title,Memr[rebwl],xlab,nsmooth)
		    call sprintf (title, SZ_LINE, "%s residuals (object)")
			call pargstr (specname)
		    }
		else {
		    call sprintf (title, SZ_LINE, "%d: %s object (residuals)")
			call pargi (ispec)
			call pargstr (specname)
		    }
		call plotspec (npout,Memr[resspec],title,Memr[rebwl],xlab,nsmooth)
		}

	    rmode = 2
	    rmode = clgeti ("report_mode")

	# Print results
	    if (debug) {
		call printf ("EIGENFIT:  %d eigenvectors:\n")
		    call pargi (nvf)
		do i = 1, nvf {
		    call printf ("EIGENFIT:  %d: %g\n")
			call pargi (i)
			call pargd (coeff[i])
		    }
#		call printf ("EIGENFIT: Res: %f %f %f\n")
#		    call pargr (Memr[resspec])
#		    call pargr (Memr[resspec+1])
#		    call pargr (Memr[resspec+2])
		call flush (STDOUT)
		}
	    }

	else {
	    call printf ("EIGENFIT: No eigenvectors:  no fit\n")
	    call flush (STDOUT)
	    }

	# Remove continuum from residuals, if requested
	if (conproc == OUTCONT) {
	    isvec = FALSE
	    call eigencon (npout,Memr[rebspec],Memr[rebwl],specname,nsmooth,
			   ispec,nspec,isvec,Memr[rebcont])

	    # Zero regions around bad lines in residuals
	    if (badproc == ZEROBAD || badproc == NZEROBAD)
		call zerolist (npspec, Memr[rebspec], Memr[rebwl], debug)
	    if (pltspec) {
		if (badproc == ZEROBAD || badproc == NZEROBAD) {
		    if (nspec > 0) {
			call sprintf (title, SZ_LINE, "%d: %s cont sub, bad lines zeroed")
			    call pargi (ispec)
			    call pargstr (specname)
			}
		    else {
			call sprintf (title, SZ_LINE, "%s cont sub, bad lines zeroed")
			    call pargstr (specname)
			}
		    }
		else {
		    call sprintf (title, SZ_LINE, "%s cont sub")
			call pargstr (specname)
		    }
		call plotspec (npspec,Memr[rebspec],title,Memr[rebwl],xlab,nsmooth)
		}
	    }

#  Write residual (object), raw, and sky spectra to file
	if (nspec < 0) {
	    call ms_open (imout, outpath, npout, 1, 3, specim)
	    call imaddi (imout, "NAXIS", 3)
	    call imaddi (imout, "NAXIS2", 1)
	    call imaddi (imout, "NAXIS3", 3)
	    call imaddi (imout, "WCSDIM", 3)
	    call imaddd (imout, "CD2_2", 1.0d0)
	    call imaddd (imout, "CD3_3", 1.0d0)
	    call imaddd (imout, "LTM2_2", 1.0d0)
	    call imaddd (imout, "LTM3_3", 1.0d0)
	    call imastr (imout, "WAT0_001", "system=image")
	    call imastr (imout, "WAT1_001", "wtype=linear label=Wavelength units=angstroms")
	    call imastr (imout, "WAT2_001", "wtype=linear")
	    call imastr (imout, "WAT3_001", "wtype=linear")
	    call imastr (imout, "BANDID1", "spectrum - scaled sky")
	    call imastr (imout, "BANDID2", "raw spectrum")
	    call imastr (imout, "BANDID3", "averaged sky")

	    call ms_write (imout, 1, 1, resspec, TY_REAL, debug)
	    call ms_write (imout, 1, 2, rebspec, TY_REAL, debug)

# Remake eigenvector to include pixels excluded for fit
	    if (tmwl1 == minwav && tmdwl == dwav && nvtemp == npout) {
		do i = 1, npout {
		    Memr[rebvec[ivec]+i-1] = Memr[vecwcs+i-1]
		    }
		}
	    else {
		call rebin (Memr[vecwcs], vecsh, tmshft,
			Memr[rebvec[ivec]], npout, minwav, dwl[ivec],pixshift0)
		}
	    call ms_write (imout, 1, 3, rebvec[1], TY_REAL, debug)
	    }

#  Write residual (object) spectrum to new file or end of current file
	else {
	    call putspec (specim, specpath, specap, specband, imout, outpath,
		     Memr[resspec], npout, ispec, nspec)
	    }

# Write coefficients to header
	if (nspec < 2) {
	    if (vband > 1)
		call imastr (imout, "SVDFILE", tempid[1,1])
	    else
		call imastr (imout, "SVDFILE", vecfile)
	    if (vband > 1)
		call pargstr (tempid[1,1])
	    else
		call pargstr (vecfile)
	    call imaddd (imout, "SVDCOEFF", coeff[1])
	    if (oldcoeff > 0.0)
		call imaddd (imout, "SVDCOEFX", oldcoeff)
	    }
	else {
	    call imastr (imout, "SVDFILE", vecfile)
	    do ivec = 1, nvf {
		call sprintf (keyword, 16, "SVDCF%03d")
		    call pargi (tmpspec[ivec])
		call imaddd (imout, KEYWORD, coeff[ivec])
#		call printf ("Coefficient %d:  vector %d: %f\n")
#		    call pargi (ivec)
#		    call pargi (tmpspec[ivec])
#		    call pargd (coeff[ivec])
		}
		
	    }

# Note continuum removal method in spectrum header
	if (ispec < 2 && conproc != NOCONT) {
	    if (function == CN_SPLINE3)
		call sprintf (tstring, SZ_LINE,
			  "Subtract spline3 order %d, %d iter -%d %d")
	    else if (function == CN_LEGENDRE)
		call sprintf (tstring, SZ_LINE,
			  "Subtract legendre order %d, %d iter -%d %d")
	    else if (function == CN_CHEBYSHEV)
		call sprintf (tstring, SZ_LINE,
			  "Subtract chebyshev order %d, %d iter -%d %d")
	    else
		call sprintf (tstring, SZ_LINE,
			  "Subtract spline1 order %d, %d iter -%d %d")
		call pargi (order)
		call pargi (niterate)
		call pargr (lowrej)
		call pargr (hirej)
	    if (addback)
		call strcat (" and add back", tstring, SZ_LINE)
	    if (conproc == OUTCONT)
		call strcat (" on output", tstring, SZ_LINE)
	    call imastr (imout, "SVDCONT", tstring);
	    }

# Close the object spectnum image
endfit_
	call close_image (specim, specsh)

	return

end

# May  4 1999	New subroutine, based on RVSAO xcfit.x
# May 27 1999	Add code to write coefficients to a file
# Jun 10 1999	Add continuum subtraction HISTORY keyword
# Jun 15 1999	Fill bad lines before fitting and removing continuum
# Jul 29 1999	Add option to fill or zero bad lines
# Jul 30 1999	Make renorm a factor, not boolean

# Jan 13 2000	Add option to not fill before removing continuum if zeroing lines
# Jan 14 2000	Process bad lines from flag set externally
# Jan 18 2000	Pass continuum buffers in contin common, not as arguments
# Jan 18 2000	Add option to subtract continuum of rebinned spectrum
# Jan 19 2000	Fill or zero bad lines before removing continuum
# Apr 25 2000	Drop unused variables
# Apr 26 2000	Add line filling to eigencon()
# May 17 2000	Change MAXCACHE to MAXTEMPS

# Jul 30 2003	Set number of output pixels and allocate buffers here

# Dec 29 2005	Add option to add back continuum after subtracting fit
# Dec 30 2005	Declare rebvec and cnvec pointer arrays in contin.com

# Jan  4 2006	Convert all references to templates to eigenvectors
# Jan  9 2006	Pass buffers as pointers, enabling size adjustment
# Mar  7 2006	Write out 3-spectrum multspec file if nspec = -1
# Mar  7 2006	Drop stack pointer sp
# Mar  9 2006	Set WCSDIM, CD, and LTM for multispec output
# Mar 17 2006	Fix plot labels
# May  2 2006	Add BANDIDi and WATi_001 keywords to multispec output header
# May  2 2006	Fix bug by keeping raw spectrum with continuum intact
# Nov 17 2006	Coefficient = 1 if fit coefficient is too large
# Nov 20 2006	Change spmean to spavg to be real not double
# Nov 20 2006	Change renorm to facnorm so it can be double, not boolean
# Nov 20 2006	Change dimension of output multispec file to nx1x3

# May 29 2007	Keep lines if keeping continuum

# May 14 2008	Update calls to rebin to include pixshift0=0 argument
# May 14 2008	Write coefficients to Starbase tab-separated table

# Feb  5 2009	Set vector flag to eigencon correctly
