# File svdfit/Src/fitvecs.x
# May 12, 2008
# By Doug Mink, after Bill Press and Mike Kurtz

include "rvsao.h"

define	MMAX	50
define	NMAX	50

procedure fitvecs (y, ndata, ev, nfit, a, chi, chisq, debug)

real	y[ARB]		# Spectrum in, residuals returned
int	ndata		# Number of pixels in spectrum
pointer ev[ARB]		# Eigenvectors to fit
int	nfit		# Number of coefficients to fit
double	a[ARB]		# Fit coefficients (returned)
double	chi		# Chi (returned)
double	chisq		# Chi-squared (return)
bool	debug		# TRUE for debugging information

double	beta[MMAX]
double	afunc[MMAX]
double	covar[MMAX*MMAX] # Covariance matrix
double	ym, wt, sum
int	i,j,k, jk, kj, i1, i2
real	resi
bool	debug1

begin
	chisq = -1.d0
	debug1 = debug

	if (debug) {
	    call printf ("FITVECS: Fitting %d %d-pixel eigenvector")
		call pargi (nfit)
		call pargi (ndata)
	    if (nfit > 1)
		call printf ("s\n")
	    else
		call printf ("\n")
	    call printf ("FITVECS: %f %f %f\n")
		call pargr (y[1])
		call pargr (y[2])
		call pargr (y[3])
	    }

	# Initialize covariance matrix and coefficient vector
	do j = 1, nfit*nfit {
	    covar[j] = 0.d0
	    }
	do j = 1, nfit {
	    beta[j] = 0.d0
	    }
	do j = 1, nfit {
	    afunc[j] = 0.d0
	    }
	chi = 0.0
	chisq = 0.0
	if (debug) {
	    call printf ("FITVECS: covariance matrix and vector initialized\n")
	    do i = 1, nfit {
		call printf ("FITVECS: vector %d: %f\n")
		    call pargi (i)
		    call pargr (Memr[ev[i]])
		call flush (STDOUT)
		}
	    }

	# Find start of actual data by skipping leading zeroes
	i1 = 0
	do i = 1, ndata {
	    if (y[i] == 0)
		i1 = i
	    else
		break
	    }

	# Find end of actual data by dropping trailing zeroes
	i2 = ndata + 1
	do i = ndata, 1, -1 {
	    if (y[i] == 0)
		i2 = i
	    else
		break
	    }

	# Return if no data to fit
	if (i2 < i1)
	    return

	do i = 1, ndata {
	    if (i > i1 && i < i2) {
		call vfit (ev, nfit, i, afunc, debug1)
		ym = y[i]
		do j = 1, nfit {
		    wt = afunc[j]
		    do k = 1,j {
			jk = j + ((k - 1) * nfit)
			covar[jk] = covar[jk] + (wt * afunc[k])
			}
		    beta[j] = beta[j] + (ym * wt)
		    }
		}
	    }
	if (nfit > 1) {
	    do j = 2, nfit {
		do k = 1, j-1 {
		    jk = j + ((k - 1) * nfit)
		    kj = k + ((j - 1) * nfit)
		    covar[kj] = covar[jk]
		    }
		}
	    }

	if (debug) {
	    call printf ("FITVECS: Covariance computed; preparing to fit\n")
	    do i = 1, nfit {
		call printf ("%d:")
		    call pargi (i)
		do j = 1, nfit {
		    call printf (" %.8g")
			call pargd (covar[j + ((i-1) * nfit)])
		    }
		call printf ("\n")
		}
	    call flush (STDOUT)
	    }

	call gaussj (covar, nfit, beta)

	if (debug) {
	    call printf ("FITVECS: Fit completed\n")
	    }

	do j = 1, nfit {
	    a[j] = beta[j]
	    }

	if (debug) {
	    call printf ("FITVECS: coeff:")
	    do j = 1, nfit {
		call printf (" %f")
		    call pargd (a[j])
		}
	    call printf ("\n")
	    call flush (STDOUT)
	    }

	chisq = 0.
	chi = 0.
	do i = 1, ndata {
	    if (i > i1 && i < i2) {
		call vfit (ev, nfit, i, afunc, debug1)
		sum = 0.
		do j = 1, nfit {
		    sum = sum + (a[j] * afunc[j])
		    }

#  chisq assumes unit variance, which cannot be true
#  sum is the model fit, note that both sum and y (the data) can
#  be zero.
		resi = double (y[i]) - sum
		chisq = chisq + (resi * resi)
		chi = chi + resi
		y[i] = real (resi)
		}
	    }

	if (debug) {
	    call printf ("FITVECS: chi^2 = %f\n")
		call pargd (chisq)
	    }

	return
end


procedure gaussj (a, n, b)

double	a[ARB]		# Correlation matrix
double	b[ARB]
int	n		# Dimension of square correlation matrix

int	ipiv[NMAX],indxr[NMAX],indxc[NMAX]
int	i, j, k, jk, irow, icol, ir, ic, ii
double	dum, pivinv, big, ajk

begin
	do j = 1, n {
	    ipiv[j] = 0
	    }

	do i = 1, n {
	    big = 0.
	    do j = 1 ,n {
		if (ipiv[j] != 1) {
		    do k = 1, n {
			if (ipiv[k] == 0) {
			    ajk = abs (a[j+((k-1)*n)])
			    if (ajk >= big) {
	 	 		big = ajk
				irow = j
				icol = k
				}
			    else if (ipiv[k] > 1) {
				call printf ("GAUSSJ: singular matrix %d=%d\n")
				    call pargi (k)
				    call pargi (ipiv[k])
				return
				}
			    }
			}
		    }
		}

	    ipiv[icol] = ipiv[icol] + 1
	    if (irow != icol) {
		do j = 1, n {
		    ir = irow + ((j-1) * n)
		    ic = icol + ((j-1) * n)
		    dum = a[ir]
		    a[ir] = a[ic]
		    a[ic] = dum
		    }

		dum = b[irow]
		b[irow] = b[icol]
		b[icol] = dum
		}
	    indxr[i] = irow
	    indxc[i] = icol
	    ii = icol + ((icol - 1) * n)
	    if (a[ii] == 0.d0) {
		call printf ("GAUSSJ: singular matrix, a[%d %d] = %f \n")
		    call pargi (icol)
		    call pargi (icol)
		    call pargd (a[ii])
		return
		}

	    pivinv = 1.d0 / a[ii]
	    a[ii] = 1.d0
	    do j = 1, n {
		ic = icol + ((j-1) * n)
		a[ic] = a[ic] * pivinv
		}

	    b[icol] = b[icol] * pivinv

	    do j = 1, n {
		if(j != icol] {
		    ic = j + ((icol - 1) * n)
		    dum = a[ic]
		    a[ic] = 0.
		    do k = 1, n {
			jk = j + ((k - 1) * n)
			ic = icol + ((k - 1) * n)
			a[jk] = a[jk] - (a[ic] * dum)
			}
		    b[j] = b[j] - (b[icol] * dum)
		    }
		}
	    }

	do j = n, 1, -1 {
	    if(indxr[j] != indxc[j]) {
		do k = 1, n {
		    ir = k + ((indxr[j] - 1) * n)
		    ic = k + ((indxc[j] - 1) * n)
		    dum = a[ir]
		    a[ir] = a[ic]
		    a[ic] = dum
		    }
		}
	    }

	return
end


procedure vfit (ev, nev, ipix, afunc, debug)

pointer	ev[nev]		# Eigenvectors to sample
int	nev		# Number of eigenvectors to sample
int	ipix		# Pixel number in spectrum and eigenvector
double	afunc[nev]	# Array of values (returned)
bool	debug		# True for diagnostic printout

int	iev, ip

begin
	ip = ipix - 1
	do iev = 1, nev {
#	    if (debug) {
#		call printf ("VFIT: %4d $2d: %f")
#		    call pargi (ipix)
#		    call pargi (iev)
#		    call pargd (afunc[iev])
#		call flush (STDOUT)
#		}
	    afunc[iev] = 0.d0
	    afunc[iev] = double (Memr[ev[iev]+ip])
	    }
	return
end

# May 25 1999	New subroutine, based on Mike Kurtz revision of Bill Press

# May 11 2000	Add diagnostic printing

# Jun  5 2002	Do not fit leading and trailing zeroes

# Dec 30 2005	Update debugging messages

# May 12 2008	Fix bug so *double* coefficients are printed in debug mode
